from typing import Any, Dict, List, Optional

from permifrost.snowflake_connector import SnowflakeConnector
from permifrost.snowflake_permission import SnowflakePermission


class SnowflakeRoleGrantChecker:
    """
    Holds permissions for roles and allows you to check if a role has a permission and also if it is able to grant a permission.
    """

    def __init__(self, conn: Optional[SnowflakeConnector] = None):
        self.conn = conn if conn is not None else SnowflakeConnector()
        # A simple in memory map to store lookups against the database.
        # In the future we can use something like https://pypi.org/project/cachetools/
        # to annotate the methods we want to cache calls to.
        self.role_permission_cache: Dict[str, Any] = {}

    def _get_permissions(self, role: str) -> Dict:
        if role not in self.role_permission_cache:
            self.role_permission_cache[
                role
            ] = self.conn.show_grants_to_role_with_grant_option(role)

        return self.role_permission_cache[role]

    def get_permissions(self, role: str) -> List[SnowflakePermission]:
        """
        Get a list of permissions that are granted to the given `role`.

        This function mainly maps the output of the SnowflakeConnector.show_grants_to_role function
        to the SnowflakePermission objects.
        """
        role_permission_dict = self._get_permissions(role)
        role_permissions = []
        for privilege, entity_types in role_permission_dict.items():
            for entity_type, entity_names in entity_types.items():
                for entity_name, options in entity_names.items():
                    name = entity_name if entity_type != "account" else "*"
                    role_permissions.append(
                        SnowflakePermission(
                            name, entity_type, [privilege], options["grant_option"]
                        )
                    )
        return role_permissions

    def _has_permission(
        self, role: Optional[str], permission: SnowflakePermission
    ) -> Optional[SnowflakePermission]:
        """
        Will return the SnowflakePermission if the <role> has the given <permission> on the <entity_name>.
        Will always return the given permission if <role> is none.
        """

        if not role:
            return permission

        role_permissions = self.get_permissions(role)

        # Always check for ownership first because it gives the most permissions
        if permission.as_owner() in role_permissions:
            # Since we don't check the grant_option of a permission when checking equality, we want to make sure
            # to return the actual permission value that was stored in the database to be most correct.
            return role_permissions[role_permissions.index(permission.as_owner())]
        if permission in role_permissions:
            return role_permissions[role_permissions.index(permission)]
        return None

    def has_permission(
        self, role: Optional[str], permission: SnowflakePermission
    ) -> bool:
        """
        Will return true if the <role> has the given <permission> on the <entity_name>.
        Where the <entity_name> is the name given in the permission object, or the "snowflaky" version. Both are checked.
        If the role has ownership of the entity in question, then this function should always return true.
        Will always return true if <role> is none.
        """
        return (
            self._has_permission(role, permission) is not None
            or self._has_permission(
                role,
                permission.with_entity_name(
                    SnowflakeConnector.snowflaky(permission.entity_name)
                ),
            )
            is not None
        )

    def _can_grant_permission(
        self, role: Optional[str], permission: SnowflakePermission
    ) -> bool:
        """
        Given the <permission>, will return true if the <role> is able to grant them.
        Will always return true if no <role> was given.
        """

        if not role:
            return True

        has_permission = self._has_permission(role, permission)

        # Ownership will let you grant any permission on that object and is always grantable.
        # Since _has_permission will return ownership over other privileges, we only need to do a single check.
        return has_permission is not None and has_permission.grant_option

    def can_grant_permission(
        self, role: Optional[str], permission: SnowflakePermission
    ) -> bool:
        """
        Given the <permission>, will return true if the <role> is able to grant them on the <entity_name>.
        Where the <entity_name> is the name given in the permission object, or the "snowflaky" version. Both are checked.
        Will always return true if no <role> was given.
        """
        return self._can_grant_permission(
            role, permission
        ) or self._can_grant_permission(
            role,
            permission.with_entity_name(
                SnowflakeConnector.snowflaky(permission.entity_name)
            ),
        )



import logging
import os
import re
from typing import Any, Dict, List, Union
from urllib.parse import quote_plus

import sqlalchemy

# To support key pair authentication
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from snowflake.sqlalchemy import URL

from permifrost.logger import GLOBAL_LOGGER as logger

# Don't show all the info log messages from Snowflake
for logger_name in ["snowflake.connector", "bot", "boto3"]:
    log = logging.getLogger(logger_name)
    log.setLevel(logging.WARNING)


class SnowflakeConnector:
    def __init__(self, config: Dict = None) -> None:
        if not config:
            config = {
                "user": os.getenv("PERMISSION_BOT_USER"),
                "password": quote_plus(os.getenv("PERMISSION_BOT_PASSWORD", "")),
                "account": os.getenv("PERMISSION_BOT_ACCOUNT"),
                "database": os.getenv("PERMISSION_BOT_DATABASE"),
                "role": os.getenv("PERMISSION_BOT_ROLE"),
                "warehouse": os.getenv("PERMISSION_BOT_WAREHOUSE"),
                "oauth_token": os.getenv("PERMISSION_BOT_OAUTH_TOKEN"),
                "key_path": os.getenv("PERMISSION_BOT_KEY_PATH"),
                "key_passphrase": os.getenv("PERMISSION_BOT_KEY_PASSPHRASE"),
                "authenticator": os.getenv("PERMISSION_BOT_AUTHENTICATOR"),
            }

        if config["oauth_token"] is not None:
            self.engine = sqlalchemy.create_engine(
                URL(
                    user=config["user"],
                    account=config["account"],
                    authenticator="oauth",
                    token=config["oauth_token"],
                    warehouse=config["warehouse"],
                )
            )
        elif config["key_path"] is not None:
            pkb = self.generate_private_key(
                config["key_path"], config.get("key_passphrase")
            )
            self.engine = sqlalchemy.create_engine(
                URL(
                    user=config["user"],
                    account=config["account"],
                    database=config["database"],
                    role=config["role"],
                    warehouse=config["warehouse"],
                ),
                connect_args={"private_key": pkb},
            )

        elif config["authenticator"] is not None:
            self.engine = sqlalchemy.create_engine(
                URL(
                    user=config["user"],
                    account=config["account"],
                    database=config["database"],
                    role=config["role"],
                    warehouse=config["warehouse"],
                    authenticator=config["authenticator"],
                ),
            )
        else:
            if not config["user"]:
                raise Exception(
                    "Validation Error: PERMISSION_BOT_USER not set. Please ensure environment variables are set."
                )
            self.engine = sqlalchemy.create_engine(
                URL(
                    user=config["user"],
                    password=config["password"],
                    account=config["account"],
                    database=config["database"],
                    role=config["role"],
                    warehouse=config["warehouse"],
                    # Enable the insecure_mode if you get OCSP errors while testing
                    # insecure_mode=True,
                )
            )

    def generate_private_key(
        self, key_path: str, key_passphrase: Union[str, None]
    ) -> bytes:
        with open(key_path, "rb") as key:
            encoded_key = None
            if key_passphrase:
                encoded_key = key_passphrase.encode()
            p_key = serialization.load_pem_private_key(
                key.read(), password=encoded_key, backend=default_backend()
            )

        return p_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

    def show_query(self, entity) -> List[str]:
        names = []

        query = f"SHOW {entity}"
        results = self.run_query(query).fetchall()

        for result in results:
            # lowercase the entity if alphanumeric plus _ else leave as is
            names.append(
                result["name"].lower()
                if bool(re.match("^[a-zA-Z0-9_]*$", result["name"]))
                else result["name"]
            )

        return names

    def show_databases(self) -> List[str]:
        return self.show_query("DATABASES")

    def show_warehouses(self) -> List[str]:
        return self.show_query("WAREHOUSES")

    def show_users(self) -> List[str]:
        return self.show_query("USERS")

    def show_schemas(self, database: str = None) -> List[str]:
        names = []

        if database:
            query = f"SHOW TERSE SCHEMAS IN DATABASE {database}"
        else:
            query = "SHOW TERSE SCHEMAS IN ACCOUNT"

        results = self.run_query(query).fetchall()

        for result in results:
            names.append(f"{result['database_name'].lower()}.{result['name'].lower()}")

        return names

    def show_tables(self, database: str = None, schema: str = None) -> List[str]:
        names = []

        if schema:
            query = f"SHOW TERSE TABLES IN SCHEMA {schema}"
        elif database:
            query = f"SHOW TERSE TABLES IN DATABASE {database}"
        else:
            query = "SHOW TERSE TABLES IN ACCOUNT"

        results = self.run_query(query).fetchall()
        for result in results:
            names.append(
                f"{result['database_name'].lower()}"
                + f".{result['schema_name'].lower()}"
                + f".{result['name'].lower()}"
            )

        return names

    def show_views(self, database: str = None, schema: str = None) -> List[str]:
        names = []

        if schema:
            query = f"SHOW TERSE VIEWS IN SCHEMA {schema}"
        elif database:
            query = f"SHOW TERSE VIEWS IN DATABASE {database}"
        else:
            query = "SHOW TERSE VIEWS IN ACCOUNT"

        results = self.run_query(query).fetchall()
        for result in results:
            names.append(
                f"{result['database_name'].lower()}"
                + f".{result['schema_name'].lower()}"
                + f".{result['name'].lower()}"
            )

        return names

    def show_future_grants(
        self, database: str = None, schema: str = None
    ) -> Dict[str, Dict[str, Dict[str, List[str]]]]:
        future_grants: Dict[str, Any] = {}

        if schema:
            query = f"SHOW FUTURE GRANTS IN SCHEMA {schema}"
        elif database:
            query = f"SHOW FUTURE GRANTS IN DATABASE {database}"
        else:
            pass

        results = self.run_query(query).fetchall()

        for result in results:
            if result["grant_to"] == "ROLE":
                role = result["grantee_name"].lower()
                privilege = result["privilege"].lower()
                granted_on = result["grant_on"].lower()

                future_grants.setdefault(role, {}).setdefault(privilege, {}).setdefault(
                    granted_on, []
                ).append(result["name"].lower())

            else:
                continue

        return future_grants

    def show_grants_to_role(self, role) -> Dict[str, Any]:
        grants: Dict[str, Any] = {}
        if role in ["*", '"*"']:
            return grants

        query = f"SHOW GRANTS TO ROLE {SnowflakeConnector.snowflaky_user_role(role)}"

        results = self.run_query(query).fetchall()

        for result in results:
            privilege = result["privilege"].lower()
            granted_on = result["granted_on"].lower()

            grants.setdefault(privilege, {}).setdefault(granted_on, []).append(
                result["name"].lower()
            )

        return grants

    def show_grants_to_role_with_grant_option(self, role) -> Dict[str, Any]:
        grants: Dict[str, Any] = {}

        query = f"SHOW GRANTS TO ROLE {SnowflakeConnector.snowflaky(role)}"
        results = self.run_query(query).fetchall()

        for result in results:
            privilege = result["privilege"].lower()
            granted_on = result["granted_on"].lower()
            grant_option = result["grant_option"].lower() == "true"
            name = result["name"].lower()

            grants.setdefault(privilege, {}).setdefault(granted_on, {}).setdefault(
                name, {}
            ).update({"grant_option": grant_option})

        return grants

    def show_roles_granted_to_user(self, user) -> List[str]:
        roles = []

        query = f"SHOW GRANTS TO USER {SnowflakeConnector.snowflaky_user_role(user)}"
        results = self.run_query(query).fetchall()

        for result in results:
            roles.append(result["role"].lower())

        return roles

    def get_current_user(self) -> str:
        query = "SELECT CURRENT_USER() AS USER"
        result = self.run_query(query).fetchone()
        return result["user"].lower()

    def get_current_role(self) -> str:
        query = "SELECT CURRENT_ROLE() AS ROLE"
        result = self.run_query(query).fetchone()
        return result["role"].lower()

    def show_roles(self) -> Dict[str, str]:
        roles = {}

        query = "SHOW ROLES"
        results = self.run_query(query).fetchall()

        for result in results:
            roles[result["name"].lower()] = result["owner"].lower()
        return roles

    def run_query(self, query: str):

        with self.engine.connect() as connection:
            logger.debug(f"Running query: {query}")
            result = connection.execute(query)

        return result

    def full_schema_list(self, schema: str) -> List[str]:
        """
        For a given schema name, get all schemas it may be referencing.

        For example, if <db>.* is given then all schemas in the database
        will be returned. If <db>.<schema_partial>_* is given, then all
        schemas that match the schema partial pattern will be returned.
        If a full schema name is given, it will return that single schema
        as a list.

        This function can be enhanced in the future to handle more
        complicated schema names if necessary.

        Returns a list of schema names.
        """
        # Generate the information_schema identifier for that database
        # in order to be able to filter it out
        name_parts = schema.split(".")

        info_schema = f"{name_parts[0]}.information_schema"

        fetched_schemas = []

        # All Schemas
        if name_parts[1] == "*":
            db_schemas = self.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                if db_schema != info_schema:
                    fetched_schemas.append(db_schema)

        # Prefix schema match
        elif "*" in name_parts[1]:
            db_schemas = self.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                schema_name = db_schema.split(".", 1)[1].lower()
                if schema_name.startswith(name_parts[1].split("*", 1)[0]):
                    fetched_schemas.append(db_schema)

        # TODO Handle more complicated matches

        else:
            # If no * in name, then return provided schema name
            fetched_schemas = [schema]

        return fetched_schemas

    @staticmethod
    def snowflaky(name: str) -> str:
        """
        Convert an entity name to an object identifier that will most probably be
        the proper name for Snowflake.

        e.g. gitlab-ci --> "gitlab-ci"
             527-INVESTIGATE$ISSUES.ANALYTICS.COUNTRY_CODES -->
             --> "527-INVESTIGATE$ISSUES".ANALYTICS.COUNTRY_CODES;

        Pronounced /snəʊfleɪkɪ/ like saying very fast snowflak[e and clarif]y
        Permission granted to use snowflaky as a verb.
        """
        name_parts = name.split(".")
        new_name_parts = []

        for part in name_parts:
            if (
                re.match("^[0-9a-zA-Z_]*$", part) is None  # Proper formatting
                and re.match('^".*"$', part) is None  # Already quoted
            ):
                new_name_parts.append(f'"{part}"')
            else:
                new_name_parts.append(part)

        return ".".join(new_name_parts)

    @staticmethod
    def snowflaky_user_role(name: str) -> str:
        """
        Convert users/roles to an object identifier that will most probably be
        the proper name for Snowflake.

        e.g. gitlab-ci --> "gitlab-ci"
             blake.enyart@gmail.com --> "blake.enyart@gmail.com"

        Pronounced /snəʊfleɪkɪ/ like saying very fast snowflak[e and clarif]y
        Permission granted to use snowflaky as a verb.
        """
        if (
            re.match("^[0-9a-zA-Z_]*$", name) is None  # Proper formatting
            and re.match('^".*"$', name) is None  # Already quoted
        ):
            name = f'"{name}"'

        return name



import sys

import click

from permifrost import SpecLoadingError
from permifrost.snowflake_connector import SnowflakeConnector
from permifrost.snowflake_spec_loader import SnowflakeSpecLoader

from . import cli


def print_command(command, diff, dry=False):
    """Prints the queries to the command line with prefixes"""
    diff_prefix = ""
    if command["already_granted"]:
        if diff:
            diff_prefix = "  "
        else:
            pass
    else:
        if diff:
            diff_prefix = "+ "

    if command.get("run_status"):
        foreground_color = "green"
        run_prefix = "[SUCCESS] "
    elif command.get("run_status") is None and dry:
        foreground_color = "cyan"
        run_prefix = "[PENDING] "
    elif command.get("run_status") is None:
        foreground_color = "cyan"
        run_prefix = "[SKIPPED] "
    else:
        foreground_color = "red"
        run_prefix = "[ERROR] "

    click.secho(f"{diff_prefix}{run_prefix}{command['sql']};", fg=foreground_color)


@cli.command()  # type: ignore
@click.argument("spec")
@click.option("--dry", help="Do not actually run, just check.", is_flag=True)
@click.option(
    "--diff", help="Show full diff, both new and existing permissions.", is_flag=True
)
@click.option(
    "--role",
    multiple=True,
    default=[],
    help="Run grants for specific roles. Usage: --role testrole --role testrole2.",
)
@click.option(
    "--user",
    multiple=True,
    default=[],
    help="Run grants for specific users. Usage: --user testuser --user testuser2.",
)
@click.option(
    "--ignore-memberships",
    help="Do not handle role membership grants/revokes",
    is_flag=True,
)
@click.pass_context
def run(ctx, spec, dry, diff, role, user, ignore_memberships, print_skipped=False):
    """
    Grant the permissions provided in the provided specification file for specific users and roles
    """
    if role and user:
        run_list = ["roles", "users"]
    elif role:
        run_list = ["roles"]
    elif user:
        run_list = ["users"]
    else:
        run_list = ["roles", "users"]
    if ctx.parent.params.get("verbose", 0) >= 1:
        print_skipped = True
    permifrost_grants(
        spec=spec,
        dry=dry,
        diff=diff,
        roles=role,
        users=user,
        run_list=run_list,
        ignore_memberships=ignore_memberships,
        print_skipped=print_skipped,
    )


@click.command()
@click.argument("spec")
@click.option(
    "--role",
    multiple=True,
    default=[],
    help="Run grants for specific roles. Usage: --role testrole --role testrole2.",
)
@click.option(
    "--user",
    multiple=True,
    default=[],
    help="Run grants for specific users. Usage: --user testuser --user testuser2.",
)
@click.option(
    "--ignore-memberships",
    help="Do not handle role membership grants/revokes",
    is_flag=True,
)
@click.option(
    "--run-list",
    multiple=True,
    default=["roles", "users"],
    help="Run grants for specific users. Usage: --user testuser --user testuser2.",
)
def spec_test(spec, role, user, ignore_memberships, run_list):
    """
    Load SnowFlake spec based on the roles.yml provided. CLI use only for confirming specifications are valid.
    """
    load_specs(spec, role, user, run_list, ignore_memberships)


def load_specs(spec, role, user, run_list, ignore_memberships):
    """
    Load specs separately.
    """
    try:
        click.secho("Confirming spec loads successfully")
        spec_loader = SnowflakeSpecLoader(
            spec,
            roles=role,
            users=user,
            run_list=run_list,
            ignore_memberships=ignore_memberships,
        )
        click.secho("Snowflake specs successfully loaded", fg="green")
    except SpecLoadingError as exc:
        for line in str(exc).splitlines():
            click.secho(line, fg="red")
        sys.exit(1)

    return spec_loader


def permifrost_grants(
    spec, dry, diff, roles, users, run_list, ignore_memberships, print_skipped
):
    """Grant the permissions provided in the provided specification file."""
    spec_loader = load_specs(
        spec,
        role=roles,
        user=users,
        run_list=run_list,
        ignore_memberships=ignore_memberships,
    )

    sql_grant_queries = spec_loader.generate_permission_queries(
        roles=roles,
        users=users,
        run_list=run_list,
        ignore_memberships=ignore_memberships,
    )

    click.secho()
    if diff:
        click.secho(
            "SQL Commands generated for given spec file (Full diff with both new and already granted commands):"
        )
    else:
        click.secho("SQL Commands generated for given spec file:")
    click.secho()

    conn = SnowflakeConnector()
    for query in sql_grant_queries:
        if not dry:
            status = None
            if not query.get("already_granted"):
                try:
                    conn.run_query(query.get("sql", ""))
                    status = True
                except Exception:
                    status = False

                ran_query = query
                ran_query["run_status"] = status
                print_command(ran_query, diff)
            # If already granted, print command
            elif print_skipped:
                print_command(query, diff)
        # If dry, print commands
        else:
            if not query.get("already_granted") or print_skipped:
                print_command(query, diff, dry=True)


cli.add_command(spec_test)  # type: ignore



from typing import Dict, List

import cerberus
import yaml

from permifrost.error import SpecLoadingError
from permifrost.spec_schemas.snowflake import (
    SNOWFLAKE_SPEC_DATABASE_SCHEMA,
    SNOWFLAKE_SPEC_ROLE_SCHEMA,
    SNOWFLAKE_SPEC_SCHEMA,
    SNOWFLAKE_SPEC_USER_SCHEMA,
    SNOWFLAKE_SPEC_WAREHOUSE_SCHEMA,
)
from permifrost.types import PermifrostSpecSchema

VALIDATION_ERR_MSG = 'Spec error: {} "{}", field "{}": {}'


def ensure_valid_schema(spec: Dict) -> List[str]:
    """
    Ensure that the provided spec has no schema errors.

    Returns a list with all the errors found.
    """
    error_messages = []

    validator = cerberus.Validator(yaml.safe_load(SNOWFLAKE_SPEC_SCHEMA))
    validator.validate(spec)
    for entity_type, err_msg in validator.errors.items():
        if isinstance(err_msg[0], str):
            error_messages.append(f"Spec error: {entity_type}: {err_msg[0]}")
            continue

        for error in err_msg[0].values():
            error_messages.append(f"Spec error: {entity_type}: {error[0]}")

    if error_messages:
        return error_messages

    schema = {
        "databases": yaml.safe_load(SNOWFLAKE_SPEC_DATABASE_SCHEMA),
        "roles": yaml.safe_load(SNOWFLAKE_SPEC_ROLE_SCHEMA),
        "users": yaml.safe_load(SNOWFLAKE_SPEC_USER_SCHEMA),
        "warehouses": yaml.safe_load(SNOWFLAKE_SPEC_WAREHOUSE_SCHEMA),
    }

    validators = {
        "databases": cerberus.Validator(schema["databases"]),
        "roles": cerberus.Validator(schema["roles"]),
        "users": cerberus.Validator(schema["users"]),
        "warehouses": cerberus.Validator(schema["warehouses"]),
    }

    entities_by_type = []
    for entity_type, entities in spec.items():
        if entities and entity_type in ["databases", "roles", "users", "warehouses"]:
            entities_by_type.append((entity_type, entities))

    for entity_type, entities in entities_by_type:
        for entity_dict in entities:
            for entity_name, config in entity_dict.items():
                validators[entity_type].validate(config)
                for field, err_msg in validators[entity_type].errors.items():
                    error_messages.append(
                        VALIDATION_ERR_MSG.format(
                            entity_type, entity_name, field, err_msg[0]
                        )
                    )

    return error_messages


def load_spec(spec_path: str) -> PermifrostSpecSchema:
    """
    Load a permissions specification from a file.

    If the file is not found or at least an error is found during validation,
    raise a SpecLoadingError with the appropriate error messages.

    Otherwise, return the valid specification as a Dictionary to be used
    in other operations.

    Raises a SpecLoadingError with all the errors found in the spec if at
    least one error is found.

    Returns the spec as a dictionary if everything is OK
    """
    try:
        with open(spec_path, "r") as stream:
            spec = yaml.safe_load(stream)
    except FileNotFoundError:
        raise SpecLoadingError(f"Spec File {spec_path} not found")

    error_messages = ensure_valid_schema(spec)
    if error_messages:
        raise SpecLoadingError("\n".join(error_messages))

    return spec



import logging

import click

import permifrost
from permifrost.logger import GLOBAL_LOGGER as logger


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.option(
    "-v", "--verbose", help="Increases log level with count, e.g -vv", count=True
)
@click.version_option(version=permifrost.__version__, prog_name="permifrost")
@click.pass_context
def cli(ctx, verbose):
    logger.setLevel(logging.WARNING)
    if verbose == 1:
        logger.setLevel(logging.INFO)
    if verbose >= 2:
        logger.setLevel(logging.DEBUG)

    ctx.ensure_object(dict)



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION") as version_file:
    version = version_file.read().strip()

requires = [
    "cerberus",
    "colorama",
    "coloredlogs",
    "click",
    "click-default-group",
    "pyyaml",
    "snowflake-connector-python[secure-local-storage]",
    "snowflake-sqlalchemy",
    "sqlalchemy",
]

dev_requires = [
    "black",
    "bumpversion",
    "changelog-cli",
    "coverage",
    "flake8",
    "isort",
    "mypy",
    "pre-commit",
    "pytest",
    "pytest-cov",
    "pytest-mock",
    "types-PyYAML",
]

setup(
    name="permifrost",
    version=version,
    author="GitLab Data Team, Meltano Team, & Contributors",
    author_email="permifrost@gitlab.com",
    description="Permifrost Permissions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/gitlab-data/permifrost",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    py_modules=["permifrost"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    setup_requires=["pytest-runner"],
    tests_require=dev_requires,
    # run `make requirements.txt` after editing
    install_requires=requires,
    extras_require={"dev": dev_requires},
    entry_points={"console_scripts": ["permifrost = permifrost.cli:main"]},
)



from typing import Any, Dict, List, Set, Tuple, TypedDict

from permifrost.error import SpecLoadingError
from permifrost.logger import GLOBAL_LOGGER as logger
from permifrost.types import PermifrostSpecSchema


class EntitySchema(TypedDict):
    databases: Set[str]
    database_refs: Set[str]
    shared_databases: Set[str]
    schema_refs: Set[str]
    table_refs: Set[str]
    roles: Set[str]
    role_refs: Set[str]
    users: Set[str]
    warehouses: Set[str]
    warehouse_refs: Set[str]
    require_owner: bool


class EntityGenerator:
    def __init__(self, spec: PermifrostSpecSchema):
        self.spec = spec
        self.entities: EntitySchema = {
            "databases": set(),
            "database_refs": set(),
            "shared_databases": set(),
            "schema_refs": set(),
            "table_refs": set(),
            "roles": set(),
            "role_refs": set(),
            "users": set(),
            "warehouses": set(),
            "warehouse_refs": set(),
            "require_owner": False,
        }
        self.error_messages: List[str] = []

    def inspect_entities(self) -> EntitySchema:
        """
        Inspect a valid spec and make sure that no logic errors exist.

        e.g. a role granted to a user not defined in roles
             or a user given access to a database not defined in databases
        """
        self.generate()
        self.error_messages.extend(self.ensure_valid_entity_names(self.entities))

        self.error_messages.extend(
            self.ensure_valid_spec_for_conditional_settings(self.entities)
        )

        self.error_messages.extend(self.ensure_valid_references(self.entities))

        if self.error_messages:
            raise SpecLoadingError("\n".join(self.error_messages))

        return self.entities

    @staticmethod
    def filter_grouped_entities_by_type(grouped_entities: List, type: str) -> List:
        """
        Takes a list of grouped entities and filters them for a particular entity_type

        For example:
            filter_grouped_entities_by_type(grouped_entities, 'role') ->
                {'accountadmin', 'demo', 'securityadmin', 'sysadmin', 'useradmin'}

        """
        filtered_entities = [
            entries for entity_type, entries in grouped_entities if entity_type == type
        ]

        # Avoid returning a nested list if there are entities
        if not filtered_entities:
            return filtered_entities
        else:
            return filtered_entities[0]

    @staticmethod
    def group_spec_by_type(spec: PermifrostSpecSchema) -> List[Tuple[str, Any]]:
        """
        Takes as input the Permifrost Spec and converts it to a grouped list
        For example:
            [('databases',
                [{'demo': {'shared': False}}]),
            ('roles',
                [
                  {'accountadmin':
                    {'warehouses': ['loading'],
                     'member_of': ['sysadmin', 'securityadmin']}},
                  {'securityadmin':
                    {'warehouses': ['loading'],
                    'member_of': ['useradmin']}},
                ]
            ...
            ]

        """
        entities_by_type = [
            (entity_type, entry)
            for entity_type, entry in spec.items()
            if entry and entity_type != "version"
        ]
        return entities_by_type

    def generate(self) -> EntitySchema:
        """
        Generate and return a dictionary with all the entities defined or
        referenced in the permissions specification file.

        The xxx_refs entities are referenced by various permissions.
        For example:
        'roles' --> All the roles defined in the spec
        'role_refs' --> All the roles referenced in a member_of permission
        'table_refs' --> All the tables referenced in read/write privileges
                         or in owns entries

        Returns a tuple (entities, error_messages) with all the entities defined
        in the spec and any errors found (e.g. a user not assigned their user role)
        """
        entities_by_type = self.group_spec_by_type(self.spec)
        self.generate_roles(
            self.filter_grouped_entities_by_type(entities_by_type, "roles")
        )
        self.generate_databases(
            self.filter_grouped_entities_by_type(entities_by_type, "databases")
        )
        self.generate_warehouses(
            self.filter_grouped_entities_by_type(entities_by_type, "warehouses")
        )
        self.generate_users(
            self.filter_grouped_entities_by_type(entities_by_type, "users")
        )

        # Filter the owner requirement and set it to True or False
        require_owner = [
            entry
            for entity_type, entry in entities_by_type
            if entity_type == "require-owner"
        ]
        self.entities["require_owner"] = require_owner == [True]

        self.generate_implicit_refs_from_schemas()
        self.generate_implicit_refs_from_tables()
        # Add implicit references to DBs and Schemas.
        #  e.g. RAW.MYSCHEMA.TABLE references also DB RAW and Schema MYSCHEMA

        return self.entities

    def generate_implicit_refs_from_schemas(self):
        """Adds implicit database refs from schemas"""
        for schema in self.entities["schema_refs"]:
            name_parts = schema.split(".")
            if name_parts[0] != "*":
                self.entities["database_refs"].add(name_parts[0])

    def generate_implicit_refs_from_tables(self):
        """Adds implicit db/schema refs from tables"""
        for table in self.entities["table_refs"]:
            name_parts = table.split(".")
            if name_parts[0] != "*":
                self.entities["database_refs"].add(name_parts[0])

            if name_parts[1] != "*":
                self.entities["schema_refs"].add(f"{name_parts[0]}.{name_parts[1]}")

    def ensure_valid_entity_names(self, entities: EntitySchema) -> List[str]:
        """
        Check that all entity names are valid.

        Returns a list with all the errors found.
        """
        error_messages = []

        for db in entities["databases"].union(entities["database_refs"]):
            name_parts = db.split(".")
            if not len(name_parts) == 1:
                error_messages.append(
                    f"Name error: Not a valid database name: {db}"
                    " (Proper definition: DB)"
                )

        for schema in entities["schema_refs"]:
            name_parts = schema.split(".")
            if (not len(name_parts) == 2) or (name_parts[0] == "*"):
                error_messages.append(
                    f"Name error: Not a valid schema name: {schema}"
                    " (Proper definition: DB.[SCHEMA | *])"
                )

        for table in entities["table_refs"]:
            name_parts = table.split(".")
            if (not len(name_parts) == 3) or (name_parts[0] == "*"):
                error_messages.append(
                    f"Name error: Not a valid table name: {table}"
                    " (Proper definition: DB.[SCHEMA | *].[TABLE | *])"
                )
            elif name_parts[1] == "*" and name_parts[2] != "*":
                error_messages.append(
                    f"Name error: Not a valid table name: {table}"
                    " (Can't have a Table name after selecting all schemas"
                    " with *: DB.SCHEMA.[TABLE | *])"
                )

        return error_messages

    def ensure_valid_references(self, entities: EntitySchema) -> List[str]:
        """
        Make sure that all references are well defined.

        Returns a list with all the errors found.
        """
        error_messages = []

        # Check that all the referenced entities are also defined
        for database in entities["database_refs"]:
            if database not in entities["databases"]:
                error_messages.append(
                    f"Reference error: Database {database} is referenced "
                    "in the spec but not defined"
                )

        for role in entities["role_refs"]:
            if role not in entities["roles"] and role != "*":
                error_messages.append(
                    f"Reference error: Role {role} is referenced in the "
                    "spec but not defined"
                )

        for warehouse in entities["warehouse_refs"]:
            if warehouse not in entities["warehouses"]:
                error_messages.append(
                    f"Reference error: Warehouse {warehouse} is referenced "
                    "in the spec but not defined"
                )

        return error_messages

    def ensure_valid_spec_for_conditional_settings(
        self, entities: EntitySchema
    ) -> List[str]:
        """
        Make sure that the spec is valid based on conditional settings such as require-owner
        """
        error_messages = []

        if entities["require_owner"]:
            error_messages.extend(self.check_entities_define_owner())

        return error_messages

    def check_entities_define_owner(self) -> List[str]:
        error_messages = []

        entities_by_type = [
            (entity_type, entry)
            for entity_type, entry in self.spec.items()
            if entry and entity_type in ["databases", "roles", "users", "warehouses"]
        ]

        for entity_type, entry in entities_by_type:
            for entity_dict in entry:  # type: ignore
                for entity_name, config in entity_dict.items():
                    if "owner" not in config.keys():
                        error_messages.append(
                            f"Spec Error: Owner not defined for {entity_type} {entity_name} and require-owner is set!"
                        )

        return error_messages

    def generate_warehouses(self, warehouse_list: List[Dict[str, Dict]]) -> None:
        for warehouse_entry in warehouse_list:
            for warehouse_name, _ in warehouse_entry.items():
                self.entities["warehouses"].add(warehouse_name)

    def generate_databases(self, db_list: List[Dict[str, Dict]]) -> None:
        for db_entry in db_list:
            for db_name, config in db_entry.items():
                self.entities["databases"].add(db_name)
                if "shared" in config:
                    if type(config["shared"]) == bool:
                        if config["shared"]:
                            self.entities["shared_databases"].add(db_name)
                    else:
                        logger.debug(
                            "`shared` for database {} must be boolean, skipping Role Reference generation.".format(
                                db_name
                            )
                        )

    def generate_member_of_roles(self, config, role_name):
        try:
            if isinstance(config["member_of"], dict):
                for member_role in config["member_of"].get("include", []):
                    self.entities["roles"].add(member_role)
                for member_role in config["member_of"].get("exclude", []):
                    self.entities["roles"].add(member_role)

            if isinstance(config["member_of"], list):
                for member_role in config["member_of"]:
                    self.entities["roles"].add(member_role)
        except KeyError:
            logger.debug(
                "`member_of` not found for role {}, skipping Role Reference generation.".format(
                    role_name
                )
            )

    def generate_warehouse_roles(self, config, role_name):
        try:
            for warehouse in config["warehouses"]:
                self.entities["warehouse_refs"].add(warehouse)
        except KeyError:
            logger.debug(
                "`warehouses` not found for role {}, skipping Warehouse Reference generation.".format(
                    role_name
                )
            )

    def generate_database_roles(self, config, role_name):
        try:
            for schema in config["privileges"]["databases"]["read"]:
                self.entities["database_refs"].add(schema)
        except KeyError:
            logger.debug(
                "`privileges.databases.read` not found for role {}, skipping Database Reference generation.".format(
                    role_name
                )
            )
        try:
            for schema in config["privileges"]["databases"]["write"]:
                self.entities["database_refs"].add(schema)
        except KeyError:
            logger.debug(
                "`privileges.databases.write` not found for role {}, skipping Database Reference generation.".format(
                    role_name
                )
            )

    def generate_read_write_database_names(self, config):
        read_databases = (
            config.get("privileges", {}).get("databases", {}).get("read", [])
        )

        write_databases = (
            config.get("privileges", {}).get("databases", {}).get("write", [])
        )
        return (read_databases, write_databases)

    def generate_schema_roles(self, config, role_name):
        read_databases, write_databases = self.generate_read_write_database_names(
            config
        )

        try:
            for schema in config["privileges"]["schemas"]["read"]:
                self.entities["schema_refs"].add(schema)
                schema_db = schema.split(".")[0]
                if schema_db not in read_databases:
                    self.error_messages.append(
                        f"Privilege Error: Database {schema_db} referenced in "
                        "schema read privileges but not in database privileges "
                        f"for role {role_name}"
                    )
        except KeyError:
            logger.debug(
                "`privileges.schemas.read` not found for role {}, skipping Schema Reference generation.".format(
                    role_name
                )
            )

        try:
            for schema in config["privileges"]["schemas"]["write"]:
                self.entities["schema_refs"].add(schema)
                schema_db = schema.split(".")[0]
                if schema_db not in write_databases + read_databases:
                    self.error_messages.append(
                        f"Privilege Error: Database {schema_db} referenced in "
                        "schema write privileges but not in database privileges "
                        f"for role {role_name}"
                    )
        except KeyError:
            logger.debug(
                "`privileges.schemas.write` not found for role {}, skipping Schema Reference generation.".format(
                    role_name
                )
            )

    def generate_table_roles(self, config, role_name):
        read_databases, write_databases = self.generate_read_write_database_names(
            config
        )

        try:
            for table in config["privileges"]["tables"]["read"]:
                self.entities["table_refs"].add(table)
                table_db = table.split(".")[0]
                if table_db not in read_databases:
                    self.error_messages.append(
                        f"Privilege Error: Database {table_db} referenced in "
                        "table read privileges but not in database privileges "
                        f"for role {role_name}"
                    )
        except KeyError:
            logger.debug(
                "`privileges.tables.read` not found for role {}, skipping Table Reference generation.".format(
                    role_name
                )
            )

        try:
            for table in config["privileges"]["tables"]["write"]:
                self.entities["table_refs"].add(table)
                table_db = table.split(".")[0]
                if table_db not in write_databases + read_databases:
                    self.error_messages.append(
                        f"Privilege Error: Database {table_db} referenced in "
                        "table write privileges but not in database privileges "
                        f"for role {role_name}"
                    )
        except KeyError:
            logger.debug(
                "`privileges.tables.write` not found for role {}, skipping Table Reference generation.".format(
                    role_name
                )
            )

    def generate_ownership_roles(self, config, role_name):
        try:
            for schema in config["owns"]["databases"]:
                self.entities["database_refs"].add(schema)
        except KeyError:
            logger.debug(
                "`owns.databases` not found for role {}, skipping Database Reference generation.".format(
                    role_name
                )
            )

        try:
            for schema in config["owns"]["schemas"]:
                self.entities["schema_refs"].add(schema)
        except KeyError:
            logger.debug(
                "`owns.schemas` not found for role {}, skipping Schema Reference generation.".format(
                    role_name
                )
            )

        try:
            for table in config["owns"]["tables"]:
                self.entities["table_refs"].add(table)
        except KeyError:
            logger.debug(
                "`owns.tables` not found for role {}, skipping Table Reference generation.".format(
                    role_name
                )
            )

    def generate_roles(self, role_list):  # noqa
        """
        Generate all of the role entities.
        Also can populate the role_refs, database_refs,
        schema_refs, table_refs & warehouse_refs
        """

        for role_entry in role_list:
            for role_name, config in role_entry.items():
                self.entities["roles"].add(role_name)
                self.generate_member_of_roles(config, role_name)
                self.generate_warehouse_roles(config, role_name)
                self.generate_database_roles(config, role_name)
                self.generate_schema_roles(config, role_name)
                self.generate_table_roles(config, role_name)
                self.generate_ownership_roles(config, role_name)

    def generate_user_fn(self, config, key, ref, user_name):
        if key in config:
            for item in config[key]:
                self.entities[ref].add(item)
        else:
            logger.debug(
                "`{}` not found for user {}, skipping Role Reference generation.".format(
                    key, user_name
                )
            )

    def generate_users(self, user_list):
        """
        Generate all of the user entities.
        Also can populate the role_refs, database_refs, schema_refs & table_refs
        """

        for user_entry in user_list:
            for user_name, config in user_entry.items():
                self.entities["users"].add(user_name)
                self.generate_user_fn(config, "member_of", "role_refs", user_name)
                self.generate_user_fn(
                    config.get("owns", {}), "database", "database_refs", user_name
                )
                self.generate_user_fn(
                    config.get("owns", {}), "schemas", "schema_refs", user_name
                )
                self.generate_user_fn(
                    config.get("owns", {}), "tables", "table_refs", user_name
                )
